'''
Contains the Method class, which represents a method of an interface
whose existence and signature is being checked against.

Created on 2011-09-21

@author: xion
'''
from pyduck.utils import Interval
import inspect


class Method(object):
    ''' Represents the method of an interface. '''

    def __init__(self, actual_method):
        ''' Initializes the object from an actual method. '''
        if not inspect.isroutine(actual_method):
            raise ValueError, "Expected method/function, got %r" % actual_method
        
        arg_desc = inspect.getargspec(actual_method)
        arg_names, varargs_name, kwargs_name, default_values = arg_desc
        default_values = default_values or []
        
        self.name = actual_method.func_name
        self.has_varargs = varargs_name is not None
        self.has_kwargs = kwargs_name is not None
        self._arguments = getattr(actual_method, '_arguments', None)  # present if @expects was used
        self._returns = getattr(actual_method, '_returns', None)    # present if @returns was used
        
        fixed_arglist = not (self.has_varargs or self.has_kwargs)
        self.max_args = len(arg_names) if fixed_arglist else None
        self.min_args = len(arg_names) - len(default_values)
        
    def is_checked(self):
        ''' Returns True if the method's arguments or return value are checked
        against a specification. This happens if @expects or @returns decorators are used. '''
        return self._arguments is not None or self._returns is not None
        
    def conforms_with(self, method):
        ''' Checks whether given method is compatible with method signature
        described by this very object.
        @note: This relation is NOT symmetrical if optional arguments (incl.
        *args and **kwargs) are concerned. '''
        if isinstance(method, Method):
            method_obj = method
        elif inspect.ismethod(method):
            method_obj = Method(method)
        elif inspect.ismethoddescriptor(method):
            return False # temporary until we get to catalog all built-in descriptors
        else:
            raise ValueError, "Expected a method, got %r" % method

        # check number of arguments
        self_arg_range = Interval(self.min_args, self.max_args)
        other_arg_range = Interval(method_obj.min_args, method_obj.max_args)
        if not other_arg_range.contains(self_arg_range):
            return False
        
        # check if other method allows additional args that we don't permit
        if method_obj.has_varargs and not self.has_varargs:
            return False
        if method_obj.has_kwargs and not self.has_kwargs:
            return False
        
        # check if argument specifications match (if they are used)
        if self._arguments:
            if not method_obj._arguments:    return False
            if not self._arguments.conforms_with(method_obj._arguments):
                return False
            
        # check if return types match (if they were specified)
        if self._returns:
            if not method_obj._returns:  return False
            if not issubclass(method_obj._returns, self._returns):
                return False
        
        return True



'''
Definition of pyduck interfaces' metaclass.

Created on 2011-09-21

@author: xion
'''
from method import Method
import inspect


class InterfaceMeta(type):
    ''' Metaclass for interfaces. '''
    
    def __new__(meta, class_name, bases, class_dict): #@NoSelf
        ''' Creates the new interface. '''
        methods = get_functions(class_dict)
        iface_dict = build_iface_dict(methods)
        
        internals = get_internals(class_dict)
        iface_dict.update(internals)
        
        return type.__new__(meta, class_name, bases, iface_dict)
    
    def __instancecheck__(iface, object_or_class): #@NoSelf
        ''' Custom instance checking. Causes isinstance() to verify interface. '''
        return implements(object_or_class, iface)
    
    def __subclasscheck__(iface, cls): #@NoSelf
        ''' Custom subclass checking. Causes issubclass() to verify interface
        (including verification of "strictness" relation for two interfaces). '''
        if isinterface(cls):
            checked_iface = cls
            return contains(checked_iface, iface)
        else:
            return implements(cls, iface)


def implements(object_or_class, interface):
    ''' Verifies whether given object or class implements specified interface. '''
    for name, method in interface.__dict__.iteritems():
        if name.startswith('_'):   continue
        member = getattr(object_or_class, name, None)
        if not (member and inspect.ismethod(member)):
            return False
        if not method.conforms_with(member):
            return False
        
    return True

def contains(checked_interface, template_interface):
    ''' Verifies whether first interface is at least as strict as the second one. '''
    for name, method in template_interface.__dict__.iteritems():
        if name.startswith('_'):   continue
        member = getattr(checked_interface, name, None)
        if not member:  return False
        if not method.conforms_with(member):
            return False
        
    return True

def isinterface(obj):
    ''' Checks whether given object is a pyduck interface. '''
    meta = getattr(obj, '__metaclass__', None)
    return meta is not None and meta == InterfaceMeta


###############################################################################
# Helper functions        

def get_functions(class_dict):
    return filter(inspect.isfunction, class_dict.itervalues())

def get_internals(class_dict):
    is_internal = lambda name: name.startswith('__') and name.endswith('__')
    return dict((k, v) for k, v in class_dict.iteritems() if is_internal(k))

def build_iface_dict(methods):
    res = {}
    for method in methods:
        res[method.func_name] = Method(method)
    return res


###############################################################################
# Interface class

class Interface(object):
    ''' Base class for interfaces. It automatically uses the interface metaclass. '''
    __metaclass__ = InterfaceMeta




'''
pyduck

@summary: Python implementation of Go-like interfaces for more robust duck typing
@author: Karol Kuczmarski "Xion" 
@license: MIT
'''
from interface import InterfaceMeta, Interface, implements, contains, isinterface
from decorators import Any, expects, returns
from wrappers import enforce
from overloading import overload



'''
Module with auxiliary code used throughout the library.

Created on 2011-09-27

@author: xion
'''
import sys


class Any(object):
    ''' Marker symbol used with argument specifications.
    Accepts any type of Python object. '''
    class __metaclass__(type):
        def __instancecheck__(cls, other): #@NoSelf
            return True
        

class ArgumentSpec(dict):
    ''' Slightly customized version of standard Python dictionary
    that stores the argument specification for a function.
    It is a mapping of argument indices and names into pyduck interfaces
    or Python types. It also handles the cases of variadic arguments and
    keywords arguments if needed.
    '''
    def __init__(self, *args, **kwargs):
        super(ArgumentSpec, self).__init__(*args, **kwargs)
        self.allows_varargs = False
        self.allows_kwargs = False
    
    def __getitem__(self, key):
        try:
            return super(ArgumentSpec, self).__getitem__(key)
        except IndexError:
            if isinstance(key, (int, long)):
                if self.allows_varargs: return Any
                raise
            elif isinstance(key, str):
                if self.allows_kwargs:  return Any
                raise
            raise TypeError, "Invalid argument index or name - must be a number or ANSI string"
            
    def __setitem__(self, key, value):
        if not (isinstance(key, (int, long)) or isinstance(key, str)):
            raise TypeError, "Invalid argument index or name - must be a number or ANSI string"
        return super(ArgumentSpec, self).__setitem__(key, value)
    
    def conforms_with(self, arg_spec):
        ''' Checks whether this argument specification conforms with the other one.
        Conformance means that the set of calls that the other spec permits is no bigger
        (in terms of inclusion) than the set calls permitted by this spec. '''
        if not isinstance(arg_spec, ArgumentSpec):
            raise TypeError, "Expected argument spec, got %s (%r)" % (type(arg_spec).__name__, arg_spec)
        
        if self.allows_varargs and not arg_spec.allows_varargs: return False
        if self.allows_kwargs and not arg_spec.allows_kwargs:   return False
        
        arg_names = filter(lambda n: isinstance(n, basestring), arg_spec.keys())
        for arg_name in arg_names:
            if arg_name.startswith('_'):    continue
            self_arg_type = self.get(arg_name)
            if self_arg_type is None:   return False
            other_arg_type = arg_spec.get(arg_name)
            if not self._is_subtype(other_arg_type, self_arg_type):
                return False
            
        return True
    
    @staticmethod
    def _is_subtype(other_type, this_type):
        ''' Generalized function for checking whether one type is a subtype
        of other one. It takes all pyduck features into account, e.g. using
        boolean predicates instead of actual types. '''
        # TODO: handle boolean predicates
        return issubclass(other_type, this_type)
    

class Interval(object):
    ''' Represents a numeric interval, closed at either side. '''
    PLUS_INFINITY = sys.maxint
    MINUS_INFINITY = -sys.maxint
    
    def __init__(self, begin = None, end = None):
        self.begin = begin if begin is not None else Interval.MINUS_INFINITY
        self.end = end if end is not None else Interval.PLUS_INFINITY
        if not (self.begin <= self.end):
            raise ValueError, "Invalid interval: %s" % self
        
    def overlaps(self, other):
        ''' Checks whether this interval and the other one overlap. '''
        self_before_other = self.end < other.begin
        other_before_self = other.end < self.begin
        return not (self_before_other or other_before_self)
    
    def contains(self, other):
        ''' Checks whether this interval contains the other one. '''
        return self.begin <= other.begin <= other.end <= self.end 
    
    def is_subset_of(self, other):
        ''' Checks whether this interval is a subset of the other one. '''
        return other.contains(self)

    def __str__(self):
        return "[%s; %s]" % (self.begin, self.end)



'''
Module with decorators that enhance the functionality of pyduck interfaces.

Created on 2011-09-23

@author: xion
'''
from pyduck.utils import ArgumentSpec, Any
import functools
import inspect
import itertools


def transfer_specs(from_func, to_func):
    ''' Utility function that transfers the specifications
    from one (decorated) function to another. Used by @expects
    and @returns decorators.
    '''
    spec_attrs = ['_arguments', '_has_self', '_returns']
    
    attrs = filter(lambda a: hasattr(from_func, a), spec_attrs)
    for attr in attrs:
        from_func_value = getattr(from_func, attr)
        setattr(to_func, attr, from_func_value)
        
    return to_func


###########################################################
# @expects function decorator

class ArgumentError(TypeError):
    ''' Exception raised when arguments passed to function
    do not agree with specification given to @expects decorator. '''
    def __init__(self, message, argument = None, expected = None):
        super(ArgumentError, self).__init__(message)
        self.argument = argument
        self.expected = expected

class ExpectedParametersDecorator(object):
    ''' @expects decorator which can be applied to functions.
    It performs a check whether function's actual parameters
    match the specification (in terms of pyduck interfaces)
    passed to the decorator.
    
    Example:
    @expects(Any, FileLikeObject)
    def write(value, dest_file):
        # ...
    '''
    def __init__(self, *args, **kwargs):
        self.arg_spec = ArgumentSpec()
        for i, arg in enumerate(args):
            self.arg_spec[i] = arg
        self.arg_spec.update(kwargs)
        self.omit_self = False  # we don't know yet what function we'll decorate
    
    def __call__(self, func):
        if not inspect.isroutine(func):
            raise TypeError, "@expects can only decorate functions"
        
        self._improve_argument_spec(func)
        
        @functools.wraps(func)
        def checked_func(*args, **kwargs):
            self._validate_arguments(args, kwargs)
            return func(*args, **kwargs)
        
        checked_func = transfer_specs(func, checked_func)
        checked_func._arguments = self.arg_spec
        checked_func._has_self = self.omit_self
        return checked_func
    
    def _improve_argument_spec(self, func):
        ''' Uses the actual arguments of function to improve
        the argument specification which was saved with the decorator
        object during its creation. The improved version should work
        regardless of how the argument was used (positionally or via keyword)
        in function call.
        '''
        arg_names, varargs_name, kwargs_name, _ = inspect.getargspec(func)
        
        first_arg_is_self = len(arg_names) > 0 and arg_names[0] == 'self'
        is_unbound_method = not inspect.ismethod(func) and first_arg_is_self
        if is_unbound_method:
            self.omit_self = True
        if first_arg_is_self:
            arg_names = arg_names[1:]   # 'self' is not part of argument spec

        if len(arg_names) > len(self.arg_spec):
            raise TypeError("Expected a function with %s argument(s), found only %s"
                            % (len(self.arg_spec), len(arg_names)))
         
        for i, arg_name in enumerate(arg_names):
            arg_type = self.arg_spec.get(i) or self.arg_spec.get(arg_name)
            self.arg_spec[i] = arg_type
            self.arg_spec[arg_name] = arg_type
            
        if varargs_name:    self.arg_spec.allows_varargs = True
        if kwargs_name:     self.arg_spec.allows_kwargs = True
        
    def _validate_arguments(self, varargs, kwargs):
        ''' Checks whether specified arguments of function conform with
        the specification which was passed earlier to the decorator. '''
        if self.omit_self:
            varargs = varargs[1:]   # omit 'self'
            
        all_args = itertools.chain(enumerate(varargs), kwargs.iteritems())
        for key, arg in all_args:
            expected = self.arg_spec[key]
            if expected is Any: continue
            if not self._is(arg, expected):
                expected_type = expected.__name__
                actual_type = type(arg).__name__
                raise ArgumentError("Invalid argument: expected %s, got %s (%r)" % (expected_type, actual_type, arg),
                                    argument = arg, expected = expected)
            
    @staticmethod
    def _is(obj, type_):
        ''' Generalized functions for checking whether an object
        is an instance of given type. It supports all necessary pyduck features
        that are relevant here, such as boolean predicates. '''
        # TODO: add support for boolean predicates
        return isinstance(obj, type_)
            

expects = ExpectedParametersDecorator


###########################################################
# @returns function decorator

class ReturnValueError(TypeError):
    ''' Exception raised when value returned by function
    does not agree with specification given to @returns decorator. '''
    def __init__(self, message, returned = None, expected = None):
        super(ReturnValueError, self).__init__(message)
        self.returned = returned
        self.expected = expected

class ReturnValueDecorator(object):
    ''' @returns decorator which can be applied to functions.
    It performs a check whether function's return value
    matches the interface/type specified in the as decorator's argument.
    
    Example:
    @returns(int)
    def maybe_throws():
        die_roll = random.randint(1, 6)
        return 1 if die_roll > 3 else "foo"
    '''
    def __init__(self, retval_spec):
        self.retval_spec = retval_spec
        
    def __call__(self, func):
        if not inspect.isroutine(func):
            raise TypeError, "@returns can only decorate functions"

        @functools.wraps(func)
        def checked_func(*args, **kwargs):
            retval = func(*args, **kwargs)
            self._validate_return_value(retval)
            
        checked_func = transfer_specs(func, checked_func)
        checked_func._returns = self.retval_spec
        return checked_func

    def _validate_return_value(self, retval):
        ''' Checks whether the specified return values conforms
        with interface/type that was passed to the decorator. '''
        if not isinstance(retval, self.retval_spec):
            expected_type = self.retval_spec.__name__
            actual_type = type(retval).__name__
            raise ReturnValueError("Invalid return value: expected %s, got %s (%r)" % (expected_type, actual_type, retval),
                                   returned = retval, expected = self.retval_spec)
        
        
returns = ReturnValueDecorator



'''
Module that contains code for the overloading mechanism
based on @expects declarations.

Created on 2011-10-10

@author: xion
'''
from pyduck.decorators import ExpectedParametersDecorator, ArgumentError
from pyduck.utils import ArgumentSpec
import inspect



def overload(first, *rest):
    ''' Constructs overloaded function using  specified callables.
    Function parts should have an argument specification attached
    by previous application of @expects decorator.
    '''
    if not first:
        raise ValueError, "No functions passed to overload()"
    
    typed_funcs = __reduce_to_function_list([first] + list(rest))
    return __create_overloaded_function(typed_funcs)

    
def __reduce_to_function_list(a_list):
    res = []
    for item in a_list:
        if __is_typed_function(item):
            res.append(item)
        elif hasattr(item, '__iter__'):
            functions_on_list = filter(__is_typed_function, item)
            if len(functions_on_list) < len(item):
                raise TypeError, "List should contain only functions"
            res.extend(item)
        else:
            raise TypeError, "Expected a function or list of functions"
        
    return res
        
def __is_typed_function(arg):
    ''' Utility function that verifies whether given object
    is a "typed function", i.e. either a function with argument
    specification attached or a tuple of function and argument specification.
    '''
    if not arg: return False
    
    if inspect.isroutine(arg):
        arg_spec = getattr(arg, '_arguments', None)
        return isinstance(arg_spec, ArgumentSpec)
    
    if isinstance(arg, tuple):
        return # True -- todo: handle tuples
    
    
def __create_overloaded_function(typed_functions):
    ''' Utility function that creates the overloaded version
    of function that incorporates all the "typed functions"
    specified by the argument.
    '''
    def overloaded_function(*args, **kwargs):
        for func in typed_functions:    # TODO: more intelligent overload resolution
            arg_spec = getattr(func, '_arguments', None)
            if not arg_spec:    continue
            
            # TODO: encapsulate this mechanism inside EPD class
            expects_decorator = ExpectedParametersDecorator()
            expects_decorator.arg_spec = arg_spec
            expects_decorator.omit_self = getattr(func, '_has_self', False)
            try:
                expects_decorator._validate_arguments(args, kwargs)
            except ArgumentError:
                continue
            
            return func(*args, **kwargs)

    return overloaded_function



#!/usr/bin/env python
'''
Setup script for the pyduck project.

Created on 2011-09-25

@author: xion
'''
from setuptools import setup, find_packages


setup(name = 'pyduck',
      version = '0.5.1',
      description = 'Utility framework for easier and more robust duck typing',
      long_description = open("README.rst").read(),
      author = 'Karol Kuczmarski "Xion"',
      author_email = "karol.kuczmarski@gmail.com",
      url = "http://xion.github.com/pyduck",
      license = "BSD",
      classifiers = [
                     'Development Status :: 4 - Beta',
                     'Intended Audience :: Developers',
                     'Intended Audience :: Information Technology',
                     'Intended Audience :: Science/Research',
                     'License :: OSI Approved :: BSD License',
                     'Operating System :: OS Independent',
                     'Programming Language :: Python',
                     'Programming Language :: Python :: 2.5',
                     'Programming Language :: Python :: 2.6',
                     'Programming Language :: Python :: 2.7',
                     'Topic :: Software Development',
                     'Topic :: Software Development :: Libraries',
                     'Topic :: Software Development :: Libraries :: Python Modules',
                     ],
      
      packages = find_packages(exclude = ['tests']),
)



'''
Module with wrappers for objects that are known to
implement pyduck interfaces.

Created on 2011-10-05

@author: xion
'''
from pyduck.interface import isinterface
from pyduck.decorators import ExpectedParametersDecorator, ReturnValueDecorator
from pyduck.method import Method


class EnforceWrapper(object):
    ''' The enforce() wrapper. It ensures that given
    object implements particular pyduck interface. Additionally,
    it also allows to  take advantage  of automatic verification
    of @expects declarations on that interface's methods - as if
    they were directly applied to object's methods instead.
    '''
    def __init__(self, iface):
        if not isinterface(iface):
            raise TypeError, "enforce() requires an Interface"
        self._interface = iface

    def __call__(self, obj):
        return self.on(obj)

    def on(self, obj):
        ''' Wraps specified object's methods in @expects/@returns decorators,
        according to those defined on interface's method.
        '''
        for method_name, method in self._interface.__dict__.iteritems():
            if not isinstance(method, Method):  continue
            if not method.is_checked():         continue
            
            obj_method = getattr(obj, method_name, None)
            if not obj_method:
                exc_tuple = (obj, method_name, self._interface.__name__)
                raise TypeError, "Object (%r) does have required method %s of interface %s" % exc_tuple
            
            # decorate with @returns
            if method._returns:
                returns_decorator = ReturnValueDecorator(method._returns)
                obj_method = returns_decorator(obj_method)
            
            # decorate with @expects
            if method._arguments:
                expects_decorator = ExpectedParametersDecorator()
                expects_decorator.arg_spec = method._arguments
                obj_method = expects_decorator(obj_method)
                
            setattr(obj, method_name, obj_method)
            
        return obj


def enforce(iface, obj = None):
    ''' Enforces particular interface on given object.
    It can be used in one of following ways:
    >>> enforce(Interface).on(obj)
    >>> enforce(Interface)(obj)
    >>> enforce(Interface, obj)
    In either case, it returns a proxy object which
    has its methods wrapped in arguments' checks,
    as defined by @expects declarations applied to corresponding
    interface methods.
    '''
    wrapper = EnforceWrapper(iface)
    if obj is None:
        return wrapper
    return wrapper.on(obj)
